import os
import re
from warnings import warn
from pyensembl.genome import Genome
from pyensembl import EnsemblRelease


class VariantCallingFormat(object):
    """
    Class for storing SV information in VCF format,
    all components are in string format
    """

    def __init__(self, chrom, pos, ref, alt):
        self.chrom = chrom
        self.pos = pos
        self.ref = ref
        self.alt = alt

    def __str__(self):
        return "%s(chrom = %s, pos = %s, ref = %s, alt = %s)" % (
            self.__class__.__name__,
            self.chrom,
            self.pos,
            self.ref,
            self.alt
        )

    def __repr__(self):
        return "%s(%s, %s, %s, %s)" % (
            self.__class__.__name__,
            self.chrom,
            self.pos,
            self.ref,
            self.alt
        )


def vcf_load(filepath):
    """
    :param filepath: the absolute path of a VCF file
    :return: a list of VariantCallingFormat objects
    """
    svvcf_list = []
    filename = os.path.basename(filepath)
    line_num = 0
    print("Loading SVs from {0}.".format(filename))
    with open(filepath, 'r') as f:
        for line in f:
            line_num += 1
            if line.startswith('#'):
                continue
            else:
                tmpline = line.rstrip().split("\t")
                chrom = tmpline[0].replace('chr', '')
                pos = tmpline[1]
                ref = tmpline[3]
                alt = tmpline[4]
                if vcf_alt_format_check(alt):
                    svvcf = VariantCallingFormat(chrom, pos, ref, alt)
                    svvcf_list.append(svvcf)
                else:
                    warn("File:\t{0}\tLine:\t{1}\nMalformed SV:\t{2}\t{3}\t{4}\t{5}".format(
                        filename, line_num, chrom, pos, ref, alt
                    ))
    if not svvcf_list:
        warn("No SV was detected in {0}.".format(filename))
    else:
        print("{0} SVs were detected.".format(len(svvcf_list)))
    return svvcf_list


def vcf_alt_format_check(alt):
    """
    :param alt: the ALT field of a VCF file
    :return: bool value indicate whether it is a legal ALT field
    """
    legal_pattern_right = re.compile(r'([ATCG]+)(\]|\[)(.+?):(\d+)(\]|\[)$')
    legal_pattern_left = re.compile(r'(\]|\[)(.+?):(\d+)(\]|\[)([ATCG]+)$')
    isright = legal_pattern_right.match(alt)
    isleft = legal_pattern_left.match(alt)
    return isright or isleft


def hla_load(filepath):
    """
    :param filepath: the absolute path of a HLA typing file
    :return: a list of HLA alleles joined by ,
    """
    hla_alleles = []
    filename = os.path.basename(filepath)
    with open(filepath, 'r') as f:
        for line in f:
            hla_allele = line.rstrip()
            if not hla_format_check(hla_allele):
                raise IOError("{0} in file {1} is not a supported HLA format.".format(hla_allele, filename))
            hla_alleles.append(hla_allele.replace('*', ''))
    return ','.join(hla_alleles)


def hla_format_check(hla_allele):
    """
    :param hla_allele: a string indicating the hla allele
    :return: whether it is in right format
    """
    legal_pattern = re.compile(r'HLA-[ABC]\*\d{2}:\d{2}$')
    return legal_pattern.match(hla_allele)


def ensembl_load(release, gtf_file, cdna_file, cache_dir):
    """
    :param release: the release number in EMSEMBL, could be custom
    :param gtf_file: the path of gtf file if release == custom
    :param cdna_file: the path of cdna file if release == custom
    :param cache_dir: directory for pyensembl downloading
    :return: a Genome class in pyensembl
    """
    if cache_dir:
        os.environ['PYENSEMBL_CACHE_DIR'] = cache_dir
    if release != 'custom':
        ensembl = EnsemblRelease(int(release))
        ensembl.download()
        ensembl.index()
    else:
        ensembl = Genome(gtf_path_or_url=gtf_file,
                         transcript_fasta_paths_or_urls=cdna_file,
                         reference_name='User-defined',
                         annotation_name='User-defined')
        ensembl.index()
    return ensembl


def get_window_range(window):
    """
    :param window: the args.window given by user
    :return: a list of window sizes
    """
    start = int(window.split('-')[0])
    end = int(window.split('-')[1])
    return list(range(start, end+1))


import re
from collections import OrderedDict
from neosv.sv_class import StructuralVariant


def get_coordinate(alt):
    """
    :param alt: the ALT field of VCF format
    :return: chrom and position of this breakpoint
    """
    coord_pattern = re.compile(r'(\]|\[)(.+?):(\d+)(\]|\[)')
    coord = coord_pattern.findall(alt)[0]
    chrom, pos = coord[1].replace('chr', ''), int(coord[2])
    return chrom, pos


def get_insertion(alt):
    """
    :param alt: the ALT field of VCF format
    :return: inserted sequence of this SV
    """
    nt_pattern = re.compile(r'[ATCG]+')
    right_pattern = re.compile(r'([ATCG]+)(\]|\[)')
    nt = nt_pattern.findall(alt)[0]
    if len(nt) == 1:
        return ''
    else:
        if right_pattern.search(alt):
            return nt[1:]
        else:
            return nt[:-1]


def sv_pattern_infer(svvcf):
    """
    :param svvcf: a SV stored as a VariantCallingFormat object
    :return: a StructuralVariant object
    """
    chrom1, pos1 = svvcf.chrom.replace('chr', ''), int(svvcf.pos)
    chrom2, pos2 = get_coordinate(svvcf.alt)
    insertion = get_insertion(svvcf.alt)
    # t[p[
    pattern1 = re.compile(r'[ATCG]+\[.+?\[')
    # t]p]
    pattern2 = re.compile(r'[ATCG]+\].+?\]')
    # ]p]t
    pattern3 = re.compile(r'\].+?\][ATCG]+')
    # [p[t
    pattern4 = re.compile(r'\[.+?\[[ATCG]+')

    if pattern1.search(svvcf.alt):
        pattern = 1
    elif pattern2.search(svvcf.alt):
        pattern = 2
    elif pattern3.search(svvcf.alt):
        pattern = 3
    else:
        pattern = 4
    return StructuralVariant(chrom1, pos1, chrom2, pos2, insertion, pattern)


def remove_duplicate(sv_list):
    """
    :param sv_list: a list of StructuralVariant objects
    :return: deduplicated sv_list, maintaining the initial order
    """
    ordered_dict = OrderedDict.fromkeys(sv_list)
    sv_list_rmdup = list(ordered_dict.keys())
    print("{0} duplicated SVs were removed".format(len(sv_list)-len(sv_list_rmdup)))
    return sv_list_rmdup



class StructuralVariant(object):
    """
    Class for storing SV information:
    1. chromosomes, format: string
    2. positions, format: integer
    2. insertion sequence, format: string
    3. sv pattern, format: integer
    NOTE: sv pattern corresponds to 4 patterns in VCF specification
    """
    def __init__(self, chrom1, pos1, chrom2, pos2, insertion, pattern):
        self.chrom1 = str(chrom1).replace('chr', '')
        self.pos1 = int(pos1)
        self.chrom2 = str(chrom2).replace('chr', '')
        self.pos2 = int(pos2)
        self.insertion = insertion
        self.pattern = int(pattern)

    def __str__(self):
        return "%s(chrom1 = %s, pos1 = %d, chrom2 = %s, pos2 = %d, insertion = '%s', pattern = %d)" % (
            self.__class__.__name__,
            self.chrom1,
            self.pos1,
            self.chrom2,
            self.pos2,
            self.insertion,
            self.pattern
        )

    def __repr__(self):
        return "%s(%s, %d, %s, %d, '%s', %d)" % (
            self.__class__.__name__,
            self.chrom1,
            self.pos1,
            self.chrom2,
            self.pos2,
            self.insertion,
            self.pattern
        )

    def __eq__(self, other):
        return self.sorted_coord == other.sorted_coord

    def __hash__(self):
        return hash(self.sorted_coord)

    @property
    def sorted_coord(self):
        """
        :return: sorted coordinates of two breakpoints, used for class identity
        """
        bp1 = self.chrom1 + '_' + str(self.pos1)
        bp2 = self.chrom2 + '_' + str(self.pos2)
        return tuple(sorted([bp1, bp2]))

    @property
    def svtype(self):
        if self.chrom1 != self.chrom2:
            return 'TRA'
        else:
            if self.pos1 < self.pos2:
                if self.pattern == 1:
                    return 'DEL'
                elif self.pattern == 2:
                    return 'h2hINV'
                elif self.pattern == 3:
                    return 'DUP'
                else:
                    return 't2tINV'
            else:
                if self.pattern == 1:
                    return 'DUP'
                elif self.pattern == 2:
                    return 'h2hINV'
                elif self.pattern == 3:
                    return 'DEL'
                else:
                    return 't2tINV'



import os
from neosv.args import create_arg_parser
from neosv.input import vcf_load, hla_load, ensembl_load, get_window_range
from neosv.sv_utils import sv_pattern_infer, remove_duplicate
from neosv.annotation_utils import sv_to_sveffect
from neosv.fusion_utils import sv_to_svfusion
from neosv.sequence_utils import set_nt_seq, set_aa_seq, generate_neoepitopes
from neosv.netMHC import netmhc_pep_prep, netmhc_run, netmhc_reload, netmhc_filter
from neosv.output import write_annot, write_fusion


def main():
    # generate the args
    args = create_arg_parser()
    # load a pyensembl Genome class
    ensembl = ensembl_load(args.release, args.gtffile, args.cdnafile, args.cachedir)
    # load vcf file and transform SVs into VariantCallingFormat class
    sv_vcfs = vcf_load(args.vcffile)
    # transform SV from VariantCallingFormat class to StructuralVariant class
    svs = [sv_pattern_infer(sv_vcf) for sv_vcf in sv_vcfs]
    # remove duplicated SVs
    svs = remove_duplicate(svs)

    # annotate SVs and write to file_anno
    sv_effects = [sv_to_sveffect(sv, ensembl, args.complete) for sv in svs]
    sv_effects_flat = [sv_effect_unit for sv_effect in sv_effects for sv_effect_unit in sv_effect]
    file_anno = os.path.join(args.outdir, args.prefix + '.anno.txt')
    write_annot(file_anno, sv_effects_flat)

    if not args.anno:
        # load the hla file and join the alleles by ,
        hla_alleles = hla_load(args.hlafile)
        # specify the lengths of neoepitopes you want to predict
        window_range = get_window_range(args.window)
        # transform StructuralVariant class to SVFusion class
        sv_fusions = [sv_to_svfusion(sv, ensembl) for sv in svs]
        # remove those empty SVFusions
        sv_fusions = [sv_fusion for sv_fusion in sv_fusions if not sv_fusion.is_empty()]

        # predict neoepitopes for each SVFusion
        for sv_fusion in sv_fusions:
            sv_fusion.nt_sequence = set_nt_seq(sv_fusion)
            sv_fusion.aa_sequence = set_aa_seq(sv_fusion)
            sv_fusion.neoepitopes = generate_neoepitopes(sv_fusion, window_range)

        # predict binding affinity using netMHC
        file_netmhc_in = os.path.join(args.outdir, args.prefix + '.net.in.txt')
        file_netmhc_out = os.path.join(args.outdir, args.prefix + '.net.out.txt')
        if sv_fusions:
            netmhc_pep_prep(file_netmhc_in, sv_fusions)
            netmhc_run(args.netmhc, file_netmhc_in, hla_alleles, file_netmhc_out)
        else:
            open(file_netmhc_in, 'w').close()
            open(file_netmhc_out, 'w').close()

        # reload and filter the neoepitopes based on netMHC result
        dict_epitope = netmhc_reload(file_netmhc_out)
        dict_epitope = netmhc_filter(dict_epitope, args.aff_cutoff, args.rank_cutoff)

        # generate the final outfile
        file_fusion = os.path.join(args.outdir, args.prefix + '.neoantigen.txt')
        write_fusion(file_fusion, sv_fusions, dict_epitope)


if __name__ == '__main__':
    main()



def write_annot(filepath, sveffects):
    """
    :param filepath: outfile for annotation
    :param sveffects: a list of sveffect classes
    :return: None
    """
    header = '\t'.join(['chrom1', 'pos1', 'function1', 'gene1', 'transcript_id1', 'strand1', 'transcript_retain1',
                        'chrom2', 'pos2', 'function2', 'gene2', 'transcript_id2', 'strand2', 'transcript_retain2',
                        'svpattern', 'svtype', 'fusion'])
    with open(filepath, 'w') as f:
        f.write(header + '\n')
        for sveffect in sveffects:
            f.write('\t'.join(sveffect.output()) + '\n')


def write_fusion(filepath, svfusions, dict_neo):
    """
    :param filepath: outfile for fusion derived neoantigen
    :param svfusions: a list of svfusion classes
    :param dict_neo: a dictionary for neoepitopes
    :return: None
    """
    with open(filepath, 'w') as f:
        header = '\t'.join(['chrom1', 'pos1', 'gene1', 'transcript_id1',
                            'chrom2', 'pos2', 'gene2', 'transcript_id2',
                            'svpattern', 'svtype', 'frameshift',
                            'neoantigen', 'allele', 'affinity', 'rank'])
        f.write(header + '\n')
        for svfusion in svfusions:
            for neoepitope in svfusion.neoepitopes:
                for allele in dict_neo[neoepitope]:
                    if dict_neo[neoepitope][allele][2] == 'PASS':
                        affinity = str(dict_neo[neoepitope][allele][0])
                        rank = str(dict_neo[neoepitope][allele][1])
                        f.write('\t'.join(['\t'.join(svfusion.output()), neoepitope, allele, affinity, rank]) + '\n')


from Bio.Seq import Seq


class CDS(object):
    def __init__(self, start, end, intact, startcodon, stopcodon):
        self.start = int(start)
        self.end = int(end)
        self.intact = intact
        self.startcodon = startcodon
        self.stopcodon = stopcodon

    def __str__(self):
        return "%s(start = %d, end = %d, intact = %r, startcodon = %r, stopcodon = %r)" % (
            self.__class__.__name__,
            self.start,
            self.end,
            self.intact,
            self.startcodon,
            self.stopcodon
        )

    def __repr__(self):
        return "%s(%d, %d, %r, %r, %r)" % (
            self.__class__.__name__,
            self.start,
            self.end,
            self.intact,
            self.startcodon,
            self.stopcodon
        )


class CDSCollection(object):
    def __init__(self, transcript, cdslist, part):
        self.transcript = transcript
        self.cdslist = cdslist
        self.part = part

    def __str__(self):
        return "%s(transcript_id = %s, transcript_name = %s, part= %s)" % (
            self.__class__.__name__,
            self.transcript.transcript_id,
            self.transcript.transcript_name,
            self.part
        )

    def __repr__(self):
        return "%s(%s, %s, %s)" % (
            self.__class__.__name__,
            self.transcript.transcript_id,
            self.transcript.transcript_name,
            self.part
        )

    @property
    def cds_range(self):
        if self.part == '5':
            start = 0
            end = start + self.cut_length - 1
        else:
            end = self.comp_length - 1
            # pyensembl coding_sequence_ranges does't not include the stop codon
            # but coding_sequence includes the stop codon, so we should minus 3
            # additionally for 3' part transcript
            start = end - self.cut_length + 1 - 3
        return start, end

    @property
    def transcript_id(self):
        """
        :return: transcript id of this collection of exons
        """
        return self.transcript.transcript_id

    @property
    def strand(self):
        return self.transcript.strand

    @property
    def gene_name(self):
        """
        :return: gene name of this collection of exons
        """
        return self.transcript.gene.gene_name

    @property
    def transcript_name(self):
        """
        :return: transcript name of this collection of exons
        """
        return self.transcript.transcript_name

    @property
    def nt_sequence(self):
        seq_start, seq_end = self.cds_range
        return self.transcript.coding_sequence[seq_start: seq_end+1]

    @property
    def cut_length(self):
        return sum([(cds.end-cds.start+1) for cds in self.cdslist])

    @property
    def comp_length(self):
        return len(self.transcript.coding_sequence)


class SVFusion(object):
    def __init__(self, sv, cdscollection_1, cdscollection_2):
        self.sv = sv
        self.cc_1 = cdscollection_1
        self.cc_2 = cdscollection_2
        self.nt_sequence = None
        self.aa_sequence = None
        self.neoepitopes = None

    def __str__(self):
        return "%s(sv = %r, cc_1 = %r, cc_2 = %r, nt_sequence = %s, aa_sequence = %s)" % (
            self.__class__.__name__,
            self.sv,
            self.cc_1,
            self.cc_2,
            self.nt_sequence,
            self.aa_sequence
        )

    def __repr__(self):
        return "%s(%r, %r, %r, %s, %s)" % (
            self.__class__.__name__,
            self.sv,
            self.cc_1,
            self.cc_2,
            self.nt_sequence,
            self.aa_sequence
        )

    def __eq__(self, other):
        return self.sv.sorted_coord == other.sv.sorted_coord

    def __hash__(self):
        return hash(self.sv.sorted_coord)

    def is_empty(self):
        if not self.cc_1 or not self.cc_2:
            return True

    @property
    def nt_sequence_ins(self):
        # the insertion sequence is for the forward strand,
        # so should be adjusted by the direction of final transcript
        if self.cc_1.strand == '+':
            return self.sv.insertion
        else:
            return str(Seq(self.sv.insertion).reverse_complement())

    @property
    def nt_sequence_cds(self):
        if self.cc_1.part == '5':
            return self.cc_1.nt_sequence + self.nt_sequence_ins + self.cc_2.nt_sequence
        else:
            return self.cc_2.nt_sequence + self.nt_sequence_ins + self.cc_1.nt_sequence

    @property
    def nt_sequence_3utr(self):
        if self.cc_1.part == '5':
            return self.cc_2.transcript.three_prime_utr_sequence
        else:
            return self.cc_1.transcript.three_prime_utr_sequence

    @property
    def frame_effect(self):
        # fixed a bug: the 5' cds collection could be empty, or less than 3 aa,
        # in this case, traditional start codon is lost and the program will search for the next start codon automatically.
        # but such prediction is of low reliability, we should annotate it and remove these fusions when necessary.
        if self.cc_1.part == '5' and not self.cc_1.nt_sequence.startswith('ATG'):
            return 'Start-loss'
        if self.cc_2.part == '5' and not self.cc_2.nt_sequence.startswith('ATG'):
            return 'Start-loss'
        if len(self.nt_sequence_cds) == 3*(len(self.aa_sequence)+1):
            return 'In-frame'
        elif len(self.nt_sequence_cds) > 3*(len(self.aa_sequence)+1):
            return 'Stop-gain'
        else:
            return 'Stop-loss'

    def output(self):
        tran_id_1 = self.cc_1.transcript_id
        gene_name_1 = self.cc_1.gene_name
        tran_id_2 = self.cc_2.transcript_id
        gene_name_2 = self.cc_2.gene_name
        chrom_1, chrom_2 = self.sv.chrom1, self.sv.chrom2
        pos_1, pos_2 = self.sv.pos1, self.sv.pos2
        return [chrom_1, str(pos_1), gene_name_1, tran_id_1,
                chrom_2, str(pos_2), gene_name_2, tran_id_2,
                str(self.sv.pattern), self.sv.svtype, self.frame_effect]


__all__ = ['sv_utils',
           'sv_class',
           'annotation_utils',
           'annotation_class',
           'fusion_utils',
           'fusion_class',
           'transcript_utils',
           'sequence_utils',
           'netMHC',
           'args',
           'input',
           'output',
           'main']


def get_longest_transcript(transcripts):
    """
    :param transcripts: a list of Transcript(pyensembl) instances
    :return: the longest transcript
    """
    transcripts = sorted(transcripts, key=lambda t: t.end-t.start)
    transcript = transcripts[-1]
    return transcript


def get_transcript(chrom, pos, ensembl, complete=True):
    """
    :param chrom: chromosome with no chr of breakpoint
    :param pos: position of breakpoint
    :param ensembl: Genome instance in pyensembl
    :param complete: only consider complete transcripts
    :return: firstly return the longest complete transcript,
             if there is no complete transcript and
             complete = False, return the longest transcript
    """
    transcripts = ensembl.transcripts_at_locus(contig=str(chrom), position=int(pos))
    transcripts_comp = [transcript for transcript in transcripts if transcript.complete]
    if transcripts_comp:
        return get_longest_transcript(transcripts_comp)
    else:
        if complete:
            return None
        else:
            if transcripts:
                return get_longest_transcript(transcripts)
            else:
                return None


def get_cds_range(transcript):
    """
    :param transcript: transcript instance in pyensembl
    :return: cds intervals of this transcript
             from 5' to 3', cds 1, cds 2, ...
             [start, end], start < end
    """
    cds_ranges = transcript.coding_sequence_position_ranges
    return cds_ranges


def get_noncds_range(transcript):
    """
    :param transcript: transcript instance in pyensembl
    :return: noncds intervals of this transcript
             from 5' to 3', noncds 1, noncds 2, ...
             [start, end], start < end
             if there is no interval, for example, no utr,
             we will add a interval (0,0)
    """
    cds_ranges = transcript.coding_sequence_position_ranges
    ncds_ranges = []
    if transcript.strand == '+':
        utr5_start = transcript.start
        utr5_end = cds_ranges[0][0] - 1
        if utr5_start > utr5_end:
            ncds_ranges.append((0, 0))
        else:
            ncds_ranges.append((utr5_start, utr5_end))
        for i in range(1, len(cds_ranges)):
            cds_start = cds_ranges[i-1][1] + 1
            cds_end = cds_ranges[i][0] - 1
            ncds_ranges.append((cds_start, cds_end))
        utr3_start = cds_ranges[-1][1] + 1
        utr3_end = transcript.end
        if utr3_start > utr3_end:
            ncds_ranges.append((0, 0))
        else:
            ncds_ranges.append((utr3_start, utr3_end))
    else:
        utr5_start = cds_ranges[0][1] + 1
        utr5_end = transcript.end
        if utr5_start > utr5_end:
            ncds_ranges.append((0, 0))
        else:
            ncds_ranges.append((utr5_start, utr5_end))
        for i in range(1, len(cds_ranges)):
            cds_start = cds_ranges[i][1] + 1
            cds_end = cds_ranges[i-1][0] - 1
            ncds_ranges.append((cds_start, cds_end))
        utr3_start = transcript.start
        utr3_end = cds_ranges[-1][0] - 1
        if utr3_start > utr3_end:
            ncds_ranges.append((0, 0))
        else:
            ncds_ranges.append((utr3_start, utr3_end))
    return ncds_ranges


def get_exon_range(transcript):
    """
    :param transcript: transcript instance in pyensembl
    :return: exon intervals of this transcript
             from 5' to 3', exon 1, exon 2, ...
             [start, end], start < end
    """
    exon_ranges = []
    for exon in transcript.exons:
        exon_ranges.append((exon.start, exon.end))
    return exon_ranges


def get_intron_range(transcript):
    """
    :param transcript: transcript instance in pyensembl
    :return: non-exon intervals of this transcript
             from 5' to 3', intron 1, intron 2, ...
             [start, end], start < end
    """
    exon_ranges = get_exon_range(transcript)
    intron_ranges = []
    if transcript.strand == '+':
        for i in range(1, len(exon_ranges)):
            intron_start = exon_ranges[i-1][1] + 1
            intron_end = exon_ranges[i][0] - 1
            intron_ranges.append((intron_start, intron_end))
    else:
        for i in range(1, len(exon_ranges)):
            intron_start = exon_ranges[i][1] + 1
            intron_end = exon_ranges[i-1][0] - 1
            intron_ranges.append((intron_start, intron_end))
    if intron_ranges:
        return intron_ranges
    else:
        return [(0, 0)]


def is_overlap(pos, intervals):
    """
    :param pos: a genomic position
    :param intervals: a list of genomic intervals
    :return: whether the pos overlaps with one of intervals
             and the index of overlapped interval
    """
    for i in range(len(intervals)):
        if intervals[i][0] <= pos <= intervals[i][1]:
            return True, i
    return False, -1



import argparse
import os
import sys


def create_arg_parser():
    parser = argparse.ArgumentParser(prog="neosv")
    parser.add_argument('-vf', '--vcf-file', dest='vcffile', metavar='VCF_FILE', required=True,
                        help='Structural variants in VCF format..')
    parser.add_argument('-hf', '--hla-file', dest='hlafile', metavar='HLA_FILE', default=None,
                        help='HLA alleles (resolution: 4 digit, e.g. HLA-A*02:01), with one allele per line.')
    parser.add_argument('-np', '--netmhc-path', dest='netmhc', metavar='NETMHC_PATH', default=None,
                        help='Absolute path to the netMHCpan execution file, please skip this parameter if netMHCpan has been added to your PATH.')
    parser.add_argument('-o', '--out', dest='outdir', metavar='OUTDIR', required=True,
                        help="Folder for all result files. A new folder will be created if it does not exist.")
    parser.add_argument('-p', '--prefix', dest='prefix', metavar='PREFIX', default='sample',
                        help="This prefix will be added to all output files.")
    parser.add_argument('-r', '--release', dest='release', metavar='RELEASE', default='75',
                        help='Which reference (ENSEMBL release) you want to use. Ensembl releases that'
                             'correspond to hg18/NCBI36, hg19/GRCh37, hg38/GRCh38 are 54, 75, 95.'
                             'If your data are from other species(custom), please download the gtf '
                             'file and the cdna file from ENSEMBL website ftp://ftp.ensembl.org/pub'
                             ' and specify them using --gtf-file and --cdna-file.')
    parser.add_argument('-gf', '--gtf-file', dest='gtffile', metavar='GTF_FILE', default=None,
                        help='GTF file for the reference.')
    parser.add_argument('-cf', '--cdna-file', dest='cdnafile', metavar='CDNA_FILE', default=None,
                        help='cDNA file for the reference.')
    parser.add_argument('-pd', '--pyensembl-cache-dir', dest='cachedir', metavar='PYEMSEMBL_CACHE_DIR', default=None,
                        help='Directory for Pyensembl cache files. If not specified, the platform-specific cache folder will be used.')
    parser.add_argument('-l', '--epitope-lengths', dest='window', metavar='EPITOPE_LENGTHS', default='8-11',
                        help='Lengths of neoepitopes to predict MHC binding. Default: 8-11.')
    parser.add_argument('-ic', '--ic50-cutoff', dest='aff_cutoff', metavar='IC50_CUTOFF', type=float, default=500,
                        help='Filter neoepitopes with IC50 (nM) above this value. Default: 500.')
    parser.add_argument('-rc', '--ranking-cutoff', dest='rank_cutoff', metavar='RANKING_CUTOFF', type=float,
                        default='2.0', help='Filter neoepitopes with rank above this value. Default: 2.')
    parser.add_argument('-ct', '--complete-transcript', dest='complete', metavar='COMPLETE_TRANSCRIPT', type=bool,
                        default=False, help='Only complete transcripts will be considered for SV annotation.')
    parser.add_argument('--anno-only', dest='anno', action='store_true', default=False,
                        help='Only annotate SV without predicting neoantigens.If this argument is added,'
                        '--hla-file is not required, and you will only get the annotation result.')

    args = parser.parse_args()

    if args.release != 'custom':
        valid_range = [str(i) for i in range(54, 107)]
        if args.release not in valid_range:
            sys.exit('Release number must be between 54 and 107.')
    elif args.release == 'custom':
        if not args.gtffile or not args.cdnafile:
            sys.exit('A custom release is used, but no gtffile and cdnafile specified.')
    else:
        sys.exit('--release must be a number between 54-107 or custom.')

    if not args.anno and not args.hlafile:
        sys.exit('No HLA file specified. The annotation-only mode could be used '
                 '(--annotation-only) if you do not have HLA information.')
    if not args.anno and not args.netmhc:
        sys.exit('No netMHC exec file specified. he annotation-only mode could be used '
                 '(--annotation-only) if you do not want to predict neoantigen.')
    if not os.path.exists(args.outdir):
        os.mkdir(args.outdir)

    return args



from neosv.annotation_class import EXON, EXONCollection, SVEffect
from neosv.transcript_utils import get_transcript, get_exon_range, get_intron_range, is_overlap


def truncate_exon(transcript, direction, pos):
    """
    :param transcript: a Transcript class of pyensembl
    :param direction: 5(upstream) or 3(downstream) of the breakpoint
    :param pos: position of the breakpoint
    :return: an EXONCollection class
    """
    exon_ranges = get_exon_range(transcript)
    intron_ranges = get_intron_range(transcript)
    exon_overlap, exon_index = is_overlap(pos, exon_ranges)
    intron_overlap, intron_index = is_overlap(pos, intron_ranges)
    if transcript.strand == '+' and direction == '5':
        if exon_overlap:
            exon_list = [EXON(exon[0],
                              exon[1],
                              True
                              ) for exon in exon_ranges[:exon_index+1]]
            exon_list[-1].end = pos
            exon_list[-1].intact = False
        else:
            exon_list = [EXON(exon[0],
                              exon[1],
                              True
                              ) for exon in exon_ranges[:intron_index+1]]
    elif transcript.strand == '+' and direction == '3':
        if exon_overlap:
            exon_list = [EXON(exon[0],
                              exon[1],
                              True
                              ) for exon in exon_ranges[exon_index:]]
            exon_list[0].start = pos
            exon_list[0].intact = False
        else:
            exon_list = [EXON(exon[0],
                              exon[1],
                              True
                              ) for exon in exon_ranges[intron_index+1:]]
    elif transcript.strand == '-' and direction == '5':
        if exon_overlap:
            exon_list = [EXON(exon[0],
                              exon[1],
                              True
                              ) for exon in exon_ranges[: exon_index+1]]
            exon_list[-1].start = pos
            exon_list[-1].intact = False
        else:
            exon_list = [EXON(exon[0],
                              exon[1],
                              True
                              ) for exon in exon_ranges[:intron_index+1]]
    else:
        if exon_overlap:
            exon_list = [EXON(exon[0],
                              exon[1],
                              True
                              ) for exon in exon_ranges[exon_index:]]
            exon_list[0].end = pos
            exon_list[0].intact = False
        else:
            exon_list = [EXON(exon[0],
                              exon[1],
                              True
                              ) for exon in exon_ranges[intron_index+1:]]
    return EXONCollection(transcript, exon_list, exon_index, intron_index, direction)


def sv_to_sveffect(sv, ensembl, complete=False):
    """
    :param sv: a StructuralVariant class
    :param ensembl: a Genome class in pyensembl
    :param complete: whether to only use complete transcripts
    :return: SVEffect class
    """
    transcript_1 = get_transcript(sv.chrom1, sv.pos1, ensembl, complete)
    transcript_2 = get_transcript(sv.chrom2, sv.pos2, ensembl, complete)
    if transcript_1 and transcript_2:
        if transcript_1.strand == '+' and transcript_2.strand == '+':
            if sv.pattern == 1:
                exoncollection_1 = truncate_exon(transcript_1, '5', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '3', sv.pos2)
                return [SVEffect(sv, exoncollection_1, exoncollection_2)]
            elif sv.pattern == 2:
                exoncollection_1 = truncate_exon(transcript_1, '5', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '5', sv.pos2)
                return [SVEffect(sv, exoncollection_1, None),
                        SVEffect(sv, None, exoncollection_2)]
            elif sv.pattern == 3:
                exoncollection_1 = truncate_exon(transcript_1, '3', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '5', sv.pos2)
                return [SVEffect(sv, exoncollection_1, exoncollection_2)]
            else:
                exoncollection_1 = truncate_exon(transcript_1, '3', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '3', sv.pos2)
                return [SVEffect(sv, exoncollection_1, None),
                        SVEffect(sv, None, exoncollection_2)]
        elif transcript_1.strand == '+' and transcript_2.strand == '-':
            if sv.pattern == 1:
                exoncollection_1 = truncate_exon(transcript_1, '5', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '5', sv.pos2)
                return [SVEffect(sv, exoncollection_1, None),
                        SVEffect(sv, None, exoncollection_2)]
            elif sv.pattern == 2:
                exoncollection_1 = truncate_exon(transcript_1, '5', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '3', sv.pos2)
                return [SVEffect(sv, exoncollection_1, exoncollection_2)]
            elif sv.pattern == 3:
                exoncollection_1 = truncate_exon(transcript_1, '3', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '3', sv.pos2)
                return [SVEffect(sv, exoncollection_1, None),
                        SVEffect(sv, None, exoncollection_2)]
            else:
                exoncollection_1 = truncate_exon(transcript_1, '3', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '5', sv.pos2)
                return [SVEffect(sv, exoncollection_1, exoncollection_2)]
        elif transcript_1.strand == '-' and transcript_2.strand == '+':
            if sv.pattern == 1:
                exoncollection_1 = truncate_exon(transcript_1, '3', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '3', sv.pos2)
                return [SVEffect(sv, exoncollection_1, None),
                        SVEffect(sv, None, exoncollection_2)]
            elif sv.pattern == 2:
                exoncollection_1 = truncate_exon(transcript_1, '3', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '5', sv.pos2)
                return [SVEffect(sv, exoncollection_1, exoncollection_2)]
            elif sv.pattern == 3:
                exoncollection_1 = truncate_exon(transcript_1, '5', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '5', sv.pos2)
                return [SVEffect(sv, exoncollection_1, None),
                        SVEffect(sv, None, exoncollection_2)]
            else:
                exoncollection_1 = truncate_exon(transcript_1, '5', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '3', sv.pos2)
                return [SVEffect(sv, exoncollection_1, exoncollection_2)]
        else:
            if sv.pattern == 1:
                exoncollection_1 = truncate_exon(transcript_1, '3', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '5', sv.pos2)
                return [SVEffect(sv, exoncollection_1, exoncollection_2)]
            elif sv.pattern == 2:
                exoncollection_1 = truncate_exon(transcript_1, '3', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '3', sv.pos2)
                return [SVEffect(sv, exoncollection_1, None),
                        SVEffect(sv, None, exoncollection_2)]
            elif sv.pattern == 3:
                exoncollection_1 = truncate_exon(transcript_1, '5', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '3', sv.pos2)
                return [SVEffect(sv, exoncollection_1, exoncollection_2)]
            else:
                exoncollection_1 = truncate_exon(transcript_1, '5', sv.pos1)
                exoncollection_2 = truncate_exon(transcript_2, '5', sv.pos2)
                return [SVEffect(sv, exoncollection_1, None),
                        SVEffect(sv, None, exoncollection_2)]
    elif transcript_1 and not transcript_2:
        if transcript_1.strand == '+':
            if sv.pattern in [1, 2]:
                exoncollection = truncate_exon(transcript_1, '5', sv.pos1)
                return [SVEffect(sv, exoncollection, None)]
            else:
                exoncollection = truncate_exon(transcript_1, '3', sv.pos1)
                return [SVEffect(sv, exoncollection, None)]
        else:
            if sv.pattern in [1, 2]:
                exoncollection = truncate_exon(transcript_1, '3', sv.pos1)
                return [SVEffect(sv, exoncollection, None)]
            else:
                exoncollection = truncate_exon(transcript_1, '5', sv.pos1)
                return [SVEffect(sv, exoncollection, None)]
    elif not transcript_1 and transcript_2:
        if transcript_2.strand == '+':
            if sv.pattern in [1, 4]:
                exoncollection = truncate_exon(transcript_2, '3', sv.pos2)
                return [SVEffect(sv, None, exoncollection)]
            else:
                exoncollection = truncate_exon(transcript_2, '5', sv.pos2)
                return [SVEffect(sv, None, exoncollection)]
        else:
            if sv.pattern in [1, 4]:
                exoncollection = truncate_exon(transcript_2, '5', sv.pos2)
                return [SVEffect(sv, None, exoncollection)]
            else:
                exoncollection = truncate_exon(transcript_2, '3', sv.pos2)
                return [SVEffect(sv, None, exoncollection)]
    else:
        return [SVEffect(sv, None, None)]



class EXON(object):
    """
    Class for the genomic coordinates of an
    exon, and whether it's an intact exon.
    NOTE: start is always smaller than end
    """
    def __init__(self, start, end, intact):
        self.start = int(start)
        self.end = int(end)
        self.intact = intact

    def __str__(self):
        return "%s(start = %d, end = %d, intact = %r)" % (
            self.__class__.__name__,
            self.start,
            self.end,
            self.intact
        )

    def __repr__(self):
        return "%s(%d, %d, %r)" % (
            self.__class__.__name__,
            self.start,
            self.end,
            self.intact
        )


class EXONCollection(object):
    """
    Class for a collection of EXONs
    transcript: Transcript class of pyensembl
    exon_list: a list of EXONs
    exon_idx: index of the exon in which the transcript end, if end in intron, -1
    intron_idx: index of the exon in which the transcript end, if end in exon, -1
    part: which part of the transcript is retained, 3 or 5 prime
    """
    def __init__(self, transcript, exon_list, exon_idx, intron_idx, part):
        self.transcript = transcript
        self.exon_list = exon_list
        self.exon_idx = exon_idx
        self.intron_idx = intron_idx
        self.part = part

    def __str__(self):
        return "%s(transcript_id = %s, transcript_name = %s, exon_idx = %d, intron_idx = %d, part= %s)" % (
            self.__class__.__name__,
            self.transcript.transcript_id,
            self.transcript.transcript_name,
            self.exon_idx,
            self.intron_idx,
            self.part
        )

    def __repr__(self):
        return "%s(%s, %s, %d, %d, %s)" % (
            self.__class__.__name__,
            self.transcript.transcript_id,
            self.transcript.transcript_name,
            self.exon_idx,
            self.intron_idx,
            self.part
        )

    @property
    def transcript_id(self):
        """
        :return: transcript id of this collection of exons
        """
        return self.transcript.transcript_id

    @property
    def gene_name(self):
        """
        :return: gene name of this collection of exons
        """
        return self.transcript.gene.gene_name

    @property
    def transcript_name(self):
        """
        :return: transcript name of this collection of exons
        """
        return self.transcript.transcript_name

    @property
    def exon_num(self):
        """
        :return: number of exons in this collection
        """
        return len(self.transcript.exons)

    @property
    def breakloc(self):
        """
        :return: which element this collection ends, exon or intron
        """
        if self.exon_idx == -1:
            return 'Intron'
        else:
            return 'Exon'

    @property
    def _transcript_retain(self):
        """
        :return: a list of the elements retained in this collection,
                 the element containing the breakend is in lower case
        """
        if self.part == '5' and self.breakloc == 'Intron':
            transcript_retain = ['E{}-I{}'.format(i+1, i+1) for i in range(0, self.intron_idx+1)]
            transcript_retain[-1] = transcript_retain[-1].replace('I', 'i')
        elif self.part == '5' and self.breakloc == 'Exon':
            transcript_retain = ['E{}-I{}'.format(i+1, i+1) for i in range(0, self.exon_idx)] + \
                                ['e{}'.format(self.exon_idx+1)]
        elif self.part == '3' and self.breakloc == 'Intron':
            transcript_retain = ['I{}-E{}'.format(i+1, i+2) for i in range(self.intron_idx, self.exon_num-1)]
            transcript_retain[0] = transcript_retain[0].replace('I', 'i')
        else:
            transcript_retain = ['e{}'.format(self.exon_idx+1)] + \
                                ['I{}-E{}'.format(i+1, i+2) for i in range(self.exon_idx, self.exon_num-1)]
        return transcript_retain

    def annot_l(self):
        """
        :return: the long format of transcript
        """
        annot = '-'.join(self._transcript_retain)
        return annot

    def annot_s(self):
        """
        :return: the short format of transcript
        """
        annot = '-'.join(self._transcript_retain)
        annot_split = annot.split('-')
        if len(annot_split) <= 2:
            return '-'.join(annot_split)
        else:
            return '-'.join([annot_split[0], annot_split[-1]])


class SVEffect(object):
    """
    Class for annotating the effect of a SV
    sv: StructuralVariant class
    ec_1 and ec_2: EXONCollection class
    """
    def __init__(self, sv, ec_1, ec_2):
        self.sv = sv
        self.ec_1 = ec_1
        self.ec_2 = ec_2

    def __str__(self):
        return "%s(sv = %r, ec_1 = %r, ec_2 = %r)" % (
            self.__class__.__name__,
            self.sv,
            self.ec_1,
            self.ec_2
        )

    def __repr__(self):
        return "%s(%r, %r, %r)" % (
            self.__class__.__name__,
            self.sv,
            self.ec_1,
            self.ec_2
        )

    @property
    def fusion(self):
        if self.ec_1 and self.ec_2:
            if self.ec_1.transcript.strand == self.ec_2.transcript.strand:
                if self.sv.pattern in [1, 3]:
                    return True
                else:
                    return False
            else:
                if self.sv.pattern in [1, 3]:
                    return False
                else:
                    return True
        else:
            return False

    def output(self, annoformat='short'):
        tran_annot_1, tran_annot_2 = 'None', 'None'
        tran_id_1, tran_id_2 = 'None', 'None'
        gene_name_1, gene_name_2 = 'None', 'None'
        loc_1, loc_2 = 'Intergenic', 'Intergenic'
        strand_1, strand_2 = 'None', 'None'
        fusion = 'No-fusion'
        if annoformat == 'long':
            if self.ec_1:
                tran_id_1 = self.ec_1.transcript_id
                tran_annot_1 = self.ec_1.annot_l()
                gene_name_1 = self.ec_1.gene_name
                loc_1 = self.ec_1.breakloc
                strand_1 = self.ec_1.transcript.strand
            if self.ec_2:
                tran_id_2 = self.ec_2.transcript_id
                tran_annot_2 = self.ec_2.annot_l()
                gene_name_2 = self.ec_2.gene_name
                loc_2 = self.ec_2.breakloc
                strand_2 = self.ec_2.transcript.strand
        else:
            if self.ec_1:
                tran_id_1 = self.ec_1.transcript_id
                tran_annot_1 = self.ec_1.annot_s()
                gene_name_1 = self.ec_1.gene_name
                loc_1 = self.ec_1.breakloc
                strand_1 = self.ec_1.transcript.strand
            if self.ec_2:
                tran_id_2 = self.ec_2.transcript_id
                tran_annot_2 = self.ec_2.annot_s()
                gene_name_2 = self.ec_2.gene_name
                loc_2 = self.ec_2.breakloc
                strand_2 = self.ec_2.transcript.strand

        chrom_1, chrom_2 = self.sv.chrom1, self.sv.chrom2
        pos_1, pos_2 = self.sv.pos1, self.sv.pos2
        if self.fusion:
            fusion = 'Fusion'
        return [chrom_1, str(pos_1), loc_1, gene_name_1, tran_id_1, strand_1, tran_annot_1,
                chrom_2, str(pos_2), loc_2, gene_name_2, tran_id_2, strand_2, tran_annot_2,
                str(self.sv.pattern), self.sv.svtype, fusion]



import os
class StructuralVariant(object):
    """
    Class for storing SV information:
    1. chromosomes, format: string
    2. positions, format: integer
    2. insertion sequence, format: string
    3. sv pattern, format: integer
    NOTE: sv pattern corresponds to 4 patterns in VCF specification
    """
    def __init__(self, chrom1, pos1, chrom2, pos2, info):
        self.chrom1 = str(chrom1).replace('chr', '')
        self.pos1 = int(pos1)
        self.chrom2 = str(chrom2).replace('chr', '')
        self.pos2 = int(pos2)
        self.info = info

    def __str__(self):
        return "%s(chrom1 = %s, pos1 = %d, chrom2 = %s, pos2 = %d)" % (
            self.__class__.__name__,
            self.chrom1,
            self.pos1,
            self.chrom2,
            self.pos2
        )

    def __repr__(self):
        return "%s(%s, %d, %s, %d)" % (
            self.__class__.__name__,
            self.chrom1,
            self.pos1,
            self.chrom2,
            self.pos2
        )

    def __eq__(self, other):
        if self.bp1_sorted[0] != other.bp1_sorted[0] or self.bp2_sorted[0] != other.bp2_sorted[0]:
            return False
        else:
            if (other.bp1_sorted[1]-100) <= self.bp1_sorted[1] <= (other.bp1_sorted[1]+100) and (other.bp2_sorted[1]-100) <= self.bp2_sorted[1] <= (other.bp2_sorted[1]+100):
                return True
            else:
                return False

    @property
    def sorted_coord(self):
        """
        :return: sorted coordinates of two breakpoints, used for class identity
        """
        bp1 = [self.chrom1, self.pos1]
        bp2 = [self.chrom2, self.pos2]
        return sorted([bp1, bp2], key = lambda bp: (bp[0], bp[1]))

    @property
    def bp1_sorted(self):
        return self.sorted_coord[0]

    @property
    def bp2_sorted(self):
        return self.sorted_coord[1]


def anno_filter(annofile):
    svlist = []
    with open(annofile, 'r') as f:
        next(f)
        for line in f:
            tmpline = line.rstrip().split('\t')
            chrom1 = tmpline[0]
            pos1 = tmpline[1]
            chrom2 = tmpline[7]
            pos2 = tmpline[8]
            sv = StructuralVariant(chrom1, pos1, chrom2, pos2, line)
            if sv not in svlist:
                svlist.append(sv)
    return svlist


def neo_filter(neofile, annolist):
    sv_coord_list = [sv.chrom1 + '_' + str(sv.pos1) + '_' + sv.chrom2 + '_' + str(sv.pos2) for sv in annolist]
    svlist = []
    with open(neofile, 'r') as f:
        next(f)
        for line in f:
            tmpline = line.rstrip().split('\t')
            chrom1 = tmpline[0]
            pos1 = tmpline[1]
            chrom2 = tmpline[4]
            pos2 = tmpline[5]
            sv_coord = chrom1 + '_' + pos1 + '_' + chrom2 + '_' + pos2
            sv = StructuralVariant(chrom1, pos1, chrom2, pos2, line)
            if sv_coord in sv_coord_list:
                svlist.append(sv)
    return svlist


def write_neofile(svlist, filepath):
    header = '\t'.join(['chrom1', 'pos1', 'gene1', 'transcript_id1',
                            'chrom2', 'pos2', 'gene2', 'transcript_id2',
                            'svpattern', 'svtype', 'frameshift',
                            'neoantigen', 'allele', 'affinity', 'rank'])
    with open(filepath, 'w') as f:
        f.write(header + '\n')
        for sv in svlist:
            f.write(sv.info)


def write_annofile(svlist, filepath):
    header = '\t'.join(['chrom1', 'pos1', 'function1', 'gene1', 'transcript_id1', 'strand1', 'transcript_retain1',
                        'chrom2', 'pos2', 'function2', 'gene2', 'transcript_id2', 'strand2', 'transcript_retain2',
                        'svpattern', 'svtype', 'fusion'])
    with open(filepath, 'w') as f:
        f.write(header + '\n')
        for sv in svlist:
            f.write(sv.info)


indir = '/home/ruibinxi_pkuhpc/lustre1/shiyang/SV_neoantigen/svanno_result/'

samples = [f for f in os.listdir(indir)]
for sample in samples:
    print(sample)
    file_anno = os.path.join(indir, sample, sample + '.anno.txt')
    file_neo = os.path.join(indir, sample, sample + '.neoantigen.txt')
    annolist = anno_filter(file_anno)
    neolist = neo_filter(file_neo, annolist)
    file_anno_filtered = os.path.join(indir, sample, sample + '.anno.filtered.txt')
    file_neo_filtered = os.path.join(indir, sample, sample + '.neoantigen.filtered.txt')
    write_annofile(annolist, file_anno_filtered)
    write_neofile(neolist, file_neo_filtered)



import subprocess
import re


def netmhc_pep_prep(filepath, svfusions):
    """
    :param svfusions: a list of svfusion classes
    :param filepath: the output file
    :return: None
    """
    print("Preparing peptides for netMHC.")
    with open(filepath, 'w') as f:
        for svfusion in svfusions:
            for neoepitope in svfusion.neoepitopes:
                f.write(neoepitope + '\n')


def netmhc_run(netmhcpath, peppath, alleles, outpath):
    """
    :param netmhcpath: absolute path of the netmhc execution file
    :param peppath: input peptides file
    :param alleles: HLA alleles separated by ,
    :param outpath: outfile
    :return: None
    """
    print("Running netMHC.")
    cmd = netmhcpath \
        + ' -a ' + alleles \
        + ' -f ' + peppath \
        + ' -inptype 1' \
        + ' -BA' \
        + ' > ' + outpath
    p = subprocess.Popen(cmd, shell=True)
    p.wait()


def netmhc_reload(resultpath):
    """
    :param resultpath: the result of netmhc
    :return: a dictionary {neoepitope: {allele: [affinity, rank, FILTER]}}
    """
    pep_dic = {}
    with open(resultpath, 'r') as f:
        for line in f:
            if line.startswith("    1  HLA"):
                tmpline = re.split(r'\s+', line)
                allele = tmpline[2]
                pep = tmpline[3]
                affinity = float(tmpline[13])
                rank = float(tmpline[14])
                if pep not in pep_dic:
                    pep_dic[pep] = {}
                pep_dic[pep][allele] = [affinity, rank, 'FILTER']
    return pep_dic


def netmhc_filter(pep_dic, aff_thre, rank_thre):
    """
    :param pep_dic: the dictionary for neoepitopes generated by netmhc_load
    :param aff_thre: threshold for binding affinity
    :param rank_thre: threshold for rank percentile
    :return: a dictionary in which passed neoepitopes are labeled PASS
    """
    print("Filtering netMHC result by binding affinity.")
    for pep in pep_dic:
        for allele in pep_dic[pep]:
            if pep_dic[pep][allele][0] < aff_thre and pep_dic[pep][allele][1] < rank_thre:
                pep_dic[pep][allele][2] = 'PASS'
    return pep_dic





from Bio.Seq import Seq


def set_aa_seq(svfusion):
    """
    :function: translate the nucelotide sequence in svfusion to amino acid sequence
    :param: svfusion: a SVFusion class
    :return: the corresponding amino acid sequence
    :NOTE: if there is a stop codon, the process will stop before translating all nucleotides
    """
    dna_seq = Seq(trim_to_3x(svfusion.nt_sequence))
    mrna_seq = dna_seq.transcribe()
    aa_seq = mrna_seq.translate(to_stop=True)
    return str(aa_seq)


def set_nt_seq(svfusion):
    """
    :function: given a svfusion, paste the nucleotides in cc_1, cc_2 and 3utr
    :param svfusion: a SVFusion class
    :return: the nucleotide sequence from start codon to the end of 3utr
    """
    return svfusion.nt_sequence_cds + svfusion.nt_sequence_3utr


def trim_to_3x(nt_sequence):
    """
    :function: trim a sequence to be divisible by 3
    :return: trimmed sequence
    """
    remainder = len(nt_sequence) % 3
    if remainder:
        return nt_sequence[: -remainder]
    else:
        return nt_sequence


def generate_neoepitopes(svfusion, window_range):
    """
    :function: cut the WT and MUT protein sequence by window_range, then get the MUT specific peptides
    :param svfusion: a SVFusion class
    :param window_range: a list, specifying the range of window size, e.g. [8,9,10,11]
    :return: a list of neopeptides for svfusion
    """
    mut_peptides = []
    for window in window_range:
        mut_peptides = mut_peptides + cut_sequence(svfusion.aa_sequence, window)
    wt_peptides = []
    for window in window_range:
        wt_peptides = wt_peptides + cut_sequence(svfusion.cc_1.transcript.protein_sequence, window)
        wt_peptides = wt_peptides + cut_sequence(svfusion.cc_2.transcript.protein_sequence, window)
    return list(set(mut_peptides)-set(wt_peptides))


def cut_sequence(aa_sequence, window):
    """
    :param aa_sequence: amino acid sequence
    :param window: the size of window, window should be smaller than len(aa_sequence)
    :return: all possible slices with length = window in this sequence
    """
    if len(aa_sequence) < window:
        return []
    else:
        return [aa_sequence[i: i+window] for i in range(len(aa_sequence)-window+1)]



from neosv.fusion_class import CDS, CDSCollection, SVFusion
from neosv.transcript_utils import get_cds_range, get_noncds_range, is_overlap, get_transcript


def truncate_cds(transcript, direction, pos):
    cds_ranges = get_cds_range(transcript)
    ncds_ranges = get_noncds_range(transcript)
    cds_overlap, cds_index = is_overlap(pos, cds_ranges)
    ncds_overlap, ncds_index = is_overlap(pos, ncds_ranges)
    if transcript.strand == '+' and direction == '5':
        if cds_overlap:
            cds_list = [CDS(cds[0],
                            cds[1],
                            True,
                            False,
                            False
                            ) for cds in cds_ranges[: cds_index+1]]
            cds_list[0].startcodon = True
            cds_list[-1].end = pos
            cds_list[-1].intact = False
        else:
            cds_list = [CDS(cds[0],
                            cds[1],
                            True,
                            False,
                            False
                            ) for cds in cds_ranges[: ncds_index]]
            if cds_list:
                cds_list[0].startcodon = True
            if ncds_index == len(cds_ranges):
                cds_list[-1].stopcodon = True
    elif transcript.strand == '+' and direction == '3':
        if cds_overlap:
            cds_list = [CDS(cds[0],
                            cds[1],
                            True,
                            False,
                            False
                            ) for cds in cds_ranges[cds_index: ]]
            cds_list[0].start = pos
            cds_list[0].intact = False
            cds_list[-1].stopcodon = True
        else:
            cds_list = [CDS(cds[0],
                            cds[1],
                            True,
                            False,
                            False
                            ) for cds in cds_ranges[ncds_index: ]]
            if cds_list:
                cds_list[-1].stopcodon = True
            if ncds_index == 0:
                cds_list[0].startcodon = True
    elif transcript.strand == '-' and direction == '5':
        if cds_overlap:
            cds_list = [CDS(cds[0],
                            cds[1],
                            True,
                            False,
                            False
                            ) for cds in cds_ranges[: cds_index+1]]
            cds_list[0].startcodon = True
            cds_list[-1].start = pos
            cds_list[-1].intact = False
        else:
            cds_list = [CDS(cds[0],
                            cds[1],
                            True,
                            False,
                            False
                            ) for cds in cds_ranges[: ncds_index]]
            if cds_list:
                cds_list[0].startcodon = True
            if ncds_index == len(cds_ranges):
                cds_list[-1].stopcodon = True
    else:
        if cds_overlap:
            cds_list = [CDS(cds[0],
                            cds[1],
                            True,
                            False,
                            False
                            ) for cds in cds_ranges[cds_index:]]
            cds_list[0].start = pos
            cds_list[0].intact = False
            cds_list[-1].stopcodon = True
        else:
            cds_list = [CDS(cds[0],
                            cds[1],
                            True,
                            False,
                            False
                            ) for cds in cds_ranges[ncds_index:]]
            if cds_list:
                cds_list[-1].stopcodon = True
            if ncds_index == 0:
                cds_list[0].startcodon = True
    return CDSCollection(transcript, cds_list, direction)


def sv_to_svfusion(sv, ensembl):
    transcript_1 = get_transcript(sv.chrom1, sv.pos1, ensembl, complete=True)
    transcript_2 = get_transcript(sv.chrom2, sv.pos2, ensembl, complete=True)
    if transcript_1 and transcript_2:
        if transcript_1.strand == '+' and transcript_2.strand == '+':
            if sv.pattern == 1:
                cdscollection_1 = truncate_cds(transcript_1, '5', sv.pos1)
                cdscollection_2 = truncate_cds(transcript_2, '3', sv.pos2)
            elif sv.pattern == 2:
                cdscollection_1 = None
                cdscollection_2 = None
            elif sv.pattern == 3:
                cdscollection_1 = truncate_cds(transcript_1, '3', sv.pos1)
                cdscollection_2 = truncate_cds(transcript_2, '5', sv.pos2)
            else:
                cdscollection_1 = None
                cdscollection_2 = None
        elif transcript_1.strand == '+' and transcript_2.strand == '-':
            if sv.pattern == 1:
                cdscollection_1 = None
                cdscollection_2 = None
            elif sv.pattern == 2:
                cdscollection_1 = truncate_cds(transcript_1, '5', sv.pos1)
                cdscollection_2 = truncate_cds(transcript_2, '3', sv.pos2)
            elif sv.pattern == 3:
                cdscollection_1 = None
                cdscollection_2 = None
            else:
                cdscollection_1 = truncate_cds(transcript_1, '3', sv.pos1)
                cdscollection_2 = truncate_cds(transcript_2, '5', sv.pos2)
        elif transcript_1.strand == '-' and transcript_2.strand == '+':
            if sv.pattern == 1:
                cdscollection_1 = None
                cdscollection_2 = None
            elif sv.pattern == 2:
                cdscollection_1 = truncate_cds(transcript_1, '3', sv.pos1)
                cdscollection_2 = truncate_cds(transcript_2, '5', sv.pos2)
            elif sv.pattern == 3:
                cdscollection_1 = None
                cdscollection_2 = None
            else:
                cdscollection_1 = truncate_cds(transcript_1, '5', sv.pos1)
                cdscollection_2 = truncate_cds(transcript_2, '3', sv.pos2)
        else:
            if sv.pattern == 1:
                cdscollection_1 = truncate_cds(transcript_1, '3', sv.pos1)
                cdscollection_2 = truncate_cds(transcript_2, '5', sv.pos2)
            elif sv.pattern == 2:
                cdscollection_1 = None
                cdscollection_2 = None
            elif sv.pattern == 3:
                cdscollection_1 = truncate_cds(transcript_1, '5', sv.pos1)
                cdscollection_2 = truncate_cds(transcript_2, '3', sv.pos2)
            else:
                cdscollection_1 = None
                cdscollection_2 = None
    else:
        cdscollection_1 = None
        cdscollection_2 = None
    return SVFusion(sv, cdscollection_1, cdscollection_2)


def svfusions_to_dict(svfusions):
    """
    :param svfusions: a list of svfusion class
    :return: a dict whose key is svfusion, value is neoepitopes
    """
    svfusion_dic = {}
    for svfusion in svfusions:
        svfusion_dic[svfusion] = svfusion.neoepitopes
    return svfusion_dic



from setuptools import setup, find_packages

VERSION = '0.0.1' 
DESCRIPTION = 'Neoantigens from Structural Variations'
LONG_DESCRIPTION = 'neosv is a workflow for annotating SVs and predicting neoantigens from SVs'
# Setting up
setup(
       # the name must match the folder name 'verysimplemodule'
        name="neosv", 
        version=VERSION,
        author="Yang Shi",
        author_email="<shiyang_bio@163.com>",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        packages=find_packages(),
        install_requires=['pyensembl>=1.9.3', 'biopython>=1.79'],
        entry_points={
            "console_scripts": [
                'neosv=neosv.main:main'
            ]
        },
        keywords=['python', 'bioinformatics', 'genomics', 'neoantigen', 'structural variation'],
        classifiers= [
            "Development Status :: 5 - Production/Stable",
            "Programming Language :: Python :: 3",
            "Operating System :: MacOS",
            "Operating System :: OS/2"
        ]
)
